open Ast
open Lazy
open Format
open Utils


(** module type S represents the input signature to the ProgPrint functor.
 * It is used to specify the differences between different implementations
 * of an HOCore program *)
module type S = sig
  type send
  type receive
  type send_group
  type receive_group
  type channel
  type var
  type var_group
  type prog
  val sep: Format.formatter -> unit -> unit
  val split: prog -> channel * var_group
  val channel_size: channel -> int
  val var_size:   var_group -> int
  val send_size: send_group -> int
  val rec_size:receive_group-> int
  val channel_fold: (string -> (send_group * receive_group) -> 'a -> 'a) ->
    channel -> 'a -> 'a
  val send_group_fold: ('a -> send -> 'a) -> 'a -> send_group -> 'a
  val rec_group_fold: ('a -> receive -> 'a) -> 'a -> receive_group -> 'a
  val var_fold: ('a -> var -> 'a) -> 'a -> var_group -> 'a
  val pretty_print_send: string -> Format.formatter -> send    -> unit
  val pretty_print_rec : string -> Format.formatter -> receive -> unit
  val pretty_print_var : Format.formatter -> var -> unit
end

(** functor ProgPrint takes a S signed module P to create a function, 
 * pretty_print_prog, used to display a program represented within P *)
module ProgPrint = functor (P:S) -> struct

  type t = P.prog

  (** print_separated_obj print_obj separator fold size fmt is a generic
   * function used to display on fmt a data structure accessible through the
   * fold function, and which individual node are representable through the
   * print_obj function.
   * separator is printed between each object, but not at the end (which is
   * the use of size, the size of the data structure)
  *)
  let print_separated_obj print_obj separator fold size fmt obj =
    let disp_obj n x =
      print_obj fmt x;
      if n = 0 then () else separator fmt ();
      n - 1 in
    ignore (fold disp_obj (size - 1) obj)

  (** disp_channel displayx all emissions and receptions o a given channel *)
  let disp_channel fmt a (s,r) n =
    let s_size = P.send_size s in
    let r_size = P.rec_size  r in
    print_separated_obj (P.pretty_print_send a) P.sep P.send_group_fold
      s_size fmt s;
    if not (s_size = 0 || r_size = 0) then P.sep fmt ();
    print_separated_obj (P.pretty_print_rec a) P.sep P.rec_group_fold r_size
      fmt r;
    if n <> 0 && not (s_size = 0 && r_size = 0) then P.sep fmt ();
    n - 1

  (** pretty_print_prog display a program according to the specifications given
   * as an argument to the functor *)
  let pretty_print_prog fmt p =
    let (m,l) = P.split p in
    let m_size = P.channel_size m in
    let l_size = P.var_size l in
    if m_size = 0 && l_size = 0 then fprintf fmt "*" else begin
      ignore (print_separated_obj P.pretty_print_var P.sep P.var_fold
                l_size fmt l);
      if m_size <> 0 && l_size <> 0 then P.sep fmt ();
      ignore (P.channel_fold (disp_channel fmt) m (m_size - 1))
      end
end



(** DispProgPrint is the ProgPrinted module used to display disp_prog *)
module rec DispProgPrint : sig
  val pretty_print_prog: Format.formatter -> Ast.disp_prog -> unit
end = ProgPrint(DispProgPrintS)

and DispProgPrintS : S = struct
  type send = Ast.disp_send
  type receive = Ast.disp_receive
  type send_group = send list
  type receive_group = receive list
  type channel = (send_group * receive_group) SMap.t
  type var = Ast.pvar
  type var_group = var list
  type prog = Ast.disp_prog
  let sep fmt _ = fprintf fmt " | @,"
  let split (Ast.DProg (m,l)) = (m,l)
  let channel_size = SMap.cardinal
  let var_size  = List.length
  let send_size = List.length
  let rec_size  = List.length
  let channel_fold = SMap.fold
  let send_group_fold = List.fold_left
  let rec_group_fold = List.fold_left
  let var_fold =  List.fold_left

  let pretty_print_send a fmt p =
    fprintf fmt "%s< @[%a@] >" a DispProgPrint.pretty_print_prog p
  let pretty_print_rec a fmt (x,p) =
    fprintf fmt "%s(%s).( @[%a@] )" a x DispProgPrint.pretty_print_prog p
  let pretty_print_var fmt x =
    fprintf fmt "%s" x
end

(** pretty_print_prog prints to a formatter an already ready to be displayed 
 * program *)
let pretty_print_prog = DispProgPrint.pretty_print_prog



(** types defined and pot_env (for potential environment) are used
 * to represent an environment during its unfolding *)
type defined = Def of prog | Undef of prog annot
and pot_env = defined option list

(** conv_env converts an environment from type env to type pot_env *)
let rec conv_env (Env e) =
  let rec aux k = function
    | x::l -> aux (fun t -> Some (Undef x)::t) l
    | [] -> k [] in
  aux (fun x -> x) e


(** cut_to_none: separates the free variables of the list from
 * the ones that already have a translation in the environment.
 * More precisely, given env = [None_0, ..., None_i, Some *, ..., Some *],
 * 0 as current and a list l of sorted indexes of env,
 * cut_to_none env 0 l = (l', remain, env', current) where
 * - l' is the list extracted from l with no index below i
 * - remain is the list extracted from l with no index strictly above i
 * - env' is the list extracted from env starting at index i + 1
 * - current is i + 1.
*)
let rec cut_to_none env current = function
  | [] -> [], [], env, current
  | (x::l as v) -> (match env with
      | []   -> assert false (* Check why this is is triggered *)
      | a::e -> if current < x then cut_to_none e (current + 1) v else
          (match a with
           | None -> let (l',l'',e,c) = cut_to_none env current l
             in (x::l', l'', e, c)
           | (Some _) -> [], v, env, current))

(** unfold is the main operation, that transforms a program in a given
 * environment to a program with no free variables left within
*)
let rec unfold (env: pot_env) (Prog (m,l)) =
  (** unfold_env: mapping function to initialize the given environment by
   * unfolding internal dependencies.
   * For instance [0 [a<_>]] becomes [a<_>]
  *)
  let unfold_env = function
    | None -> None
    | Some (Def _) as x -> x
    | Some (Undef (m, e)) -> Some (Def (unfold (conv_env e) m)) in
  let env = List.map unfold_env env in
  let (l, remain, env', current) = cut_to_none env 0 l in
  (** select is the reverse conversion from a pot_env to a prog *)
  let select = function
    | None | Some (Undef _) -> assert false
    | Some (Def p) -> p in
  let m' = assemble_aux select [] env' current remain in
  let new_env = None::env in
  let assoc (s,r) =
    List.map (unfold env) s, List.map (unfold new_env) r in
  List.fold_left merge (Prog (SMap.map assoc m, l)) m'

(** module DispTopPrint is the ProgPrinted module used to display toplevels *)
module DispTopPrint = ProgPrint(struct
  type send = Ast.send Ast.annot
  type receive = Ast.receive Ast.annot
  type send_group = send list
  type receive_group = receive list
  type channel = (send_group * receive_group) SMap.t
  type var = unit
  type var_group = unit
  type prog = Ast.toplevel
  let sep fmt _ = fprintf fmt " |@."
  let split t = (t,())
  let channel_size = SMap.cardinal
  let var_size  = fun _ -> 0
  let send_size = List.length
  let rec_size  = List.length
  let channel_fold = SMap.fold
  let send_group_fold = List.fold_left
  let rec_group_fold = List.fold_left
  let var_fold =  fun _ x _ -> x

  let (Wait (x, f)) = variables

  let pretty_print_send a fmt (p,e) =
    let env = unfold (conv_env e) p in
    fprintf fmt "%s< @[%a@] >" a pretty_print_prog
      (instantiate [] env variables)
  let pretty_print_rec a fmt (p,e) =
    let env = unfold (None::(conv_env e)) p in
    let disp = instantiate [x] env (force f) in
    fprintf fmt "%s(%s).@[( %a )@]" a x pretty_print_prog disp
  let pretty_print_var fmt _ =
    fprintf fmt ""
end)

(** pretty_print_top prints to a formatter a given toplevel *)
let pretty_print_top = DispTopPrint.pretty_print_prog



(** RealProgPrint is the ProgPrinted module used to display prog,
 * without any conversion from De Bruijn to real programs *)
module rec RealProgPrint : sig
  val pretty_print_prog: Format.formatter -> Ast.prog -> unit
end = ProgPrint(RealProgPrintS)

and RealProgPrintS : S = struct
  type send = Ast.send
  type receive = Ast.receive
  type send_group = send list
  type receive_group = receive list
  type channel = (send_group * receive_group) SMap.t
  type var = Ast.var
  type var_group = var list
  type prog = Ast.prog
  let sep fmt _ = fprintf fmt " | @,"
  let split (Ast.Prog (m,l)) = (m,l)
  let channel_size = SMap.cardinal
  let var_size  = List.length
  let send_size = List.length
  let rec_size  = List.length
  let channel_fold = SMap.fold
  let send_group_fold = List.fold_left
  let rec_group_fold = List.fold_left
  let var_fold =  List.fold_left

  let pretty_print_send a fmt p =
    fprintf fmt "%s< @[%a@] >" a RealProgPrint.pretty_print_prog p
  let pretty_print_rec a fmt p =
    fprintf fmt "%s.( @[%a@] )" a RealProgPrint.pretty_print_prog p
  let pretty_print_var fmt x =
    fprintf fmt "%d" x
end

let pretty_print_real_prog = RealProgPrint.pretty_print_prog


(** print_separated_list call_function separator next is a generic
 * function used to display a list of 'a objects, which may individually 
 * be printed with call_function.
 * separator is printed between each object, and may be printed at the end
 * of the list if condition next holds.
*)
let rec print_separated_list call_function separator next fmt = function
  | []   -> fprintf fmt ""
  | [x]  -> if next then fprintf fmt "%a%a" call_function x separator ()
    else fprintf fmt "%a" call_function x
  | x::l -> fprintf fmt "%a%a%a" call_function x separator ()
              (print_separated_list call_function separator next) l


let rec pretty_print_env fmt (Env e) =
  fprintf fmt "@[[ %a ]@]" (print_separated_list pretty_print_prog_annot
                            (fun fmt _ -> fprintf fmt " ; @,") false) e


and pretty_print_prog_annot fmt (p, e) =
  fprintf fmt "%a %a" pretty_print_real_prog p pretty_print_env e  

let pretty_print_real_send a fmt (p,e) =
  fprintf fmt "%s< @[%a@] > %a" a pretty_print_real_prog p pretty_print_env e
let pretty_print_real_rec a fmt (p,e) =
  fprintf fmt "%s.@[( %a )@] %a" a pretty_print_real_prog p pretty_print_env e


module RealTopPrint = ProgPrint(struct
  type send = Ast.send Ast.annot
  type receive = Ast.receive Ast.annot
  type send_group = send list
  type receive_group = receive list
  type channel = (send_group * receive_group) SMap.t
  type var = unit
  type var_group = unit
  type prog = Ast.toplevel
  let sep fmt _ = fprintf fmt " |@."
  let split t = (t,())
  let channel_size = channel_size
  let var_size  = fun _ -> 0
  let send_size = List.length
  let rec_size  = List.length
  let channel_fold = channel_fold
  let send_group_fold = List.fold_left
  let rec_group_fold = List.fold_left
  let var_fold =  fun _ x _ -> x

  let pretty_print_send = pretty_print_real_send
  let pretty_print_rec = pretty_print_real_rec 
  let pretty_print_var fmt _ =
    fprintf fmt ""
end)

(** pretty_print_real_progtop prints to a formatter a given toplevel, with all
 * variables maintained with the de Bruijn notation *)
let pretty_print_real_top = RealTopPrint.pretty_print_prog


let pretty_print_send_annot fmt (p,e) =
  let env = unfold (conv_env e) p in
  fprintf fmt "@[%a@]" pretty_print_prog
    (instantiate [] env variables)
let pretty_print_rec_annot fmt (p,e) =
  let (Wait (x,f)) = variables in
  let env = unfold (None::(conv_env e)) p in
  let disp = instantiate [x] env (force f) in
  fprintf fmt "(%s).@[( %a )@]" x pretty_print_prog disp

(** display prints to the standard output the given toplevel *)
let display p =
  Format.printf "%a" pretty_print_top p

let display_one p =
  display p; printf "@. - - - - - - @."

let display_real p =
  Format.printf "%a" pretty_print_real_top p

